﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Net;
using System.Net.Cache;
using System.Windows.Forms;

namespace ARClient {
	partial class frmSync : Form {
		public Form1 MainForm { get; set; }
		public IEnumerable<String> Hosts { get; set; }
		private int Received = 0, Imported = 0, Exported = 0;
		private int TotalReceived = 0, TotalImported = 0, TotalExported = 0;
		private int BytesReceived = 0, BytesSent = 0;
		private int TotalBytesReceived = 0, TotalBytesSent = 0;

		public frmSync() {
			InitializeComponent();
		}

		private Byte[] RealRead(Stream from, int count, Boolean throweof) {
			Byte[] ret = new Byte[count];
			int offset = 0;
			while (count > 0) {
				int got = from.Read(ret, offset, count);
				if (backgroundWorker1.CancellationPending) break;
				if (got == 0 && offset == 0) {
					if (throweof) throw new EndOfStreamException();
					return null;
				} else if (got <= 0) {
					throw new EndOfStreamException();
				}
				count -= got;
				offset += got;
			}
			return ret;
		}

		private void backgroundWorker1_DoWork(object sender, DoWorkEventArgs e) {
			foreach (String Host in Hosts) {
				Received = Imported = Exported = BytesSent = BytesReceived = 0;
				if (backgroundWorker1.CancellationPending) break;
				String url = Host;
				int method = 3;
				if (Host.StartsWith("get:", StringComparison.InvariantCultureIgnoreCase)) {
					url = Host.Substring(4);
					method = 1;
				} else if (Host.StartsWith("post:", StringComparison.InvariantCultureIgnoreCase)) {
					url = Host.Substring(5);
					method = 2;
				} else if (Host.StartsWith("put:", StringComparison.InvariantCultureIgnoreCase)) {
					url = Host.Substring(3);
					method = 3;
				}
				try {
					backgroundWorker1.ReportProgress(0, "Preparing...");
					backgroundWorker1.ReportProgress(1, Host);
					if (method != 1 && method != 3) throw new NotImplementedException("The requested method is not implemented");
					String args = String.Format("version={0}&random={1}&get={2}", 3, (new Random()).Next(), 0);
					url += (url.Contains("?") ? "&" : "?") + args;
					WebRequest request = WebRequest.Create(url);
					try { request.CachePolicy = new RequestCachePolicy(RequestCacheLevel.BypassCache); } catch { }
					if (method == 3) {
						request.Method = "PUT";
					} else {
						request.Method = "GET";
					}
					request.Timeout = 30000;
					backgroundWorker1.ReportProgress(0, "Preparing data...");
					if (method == 3) {
						using (Stream stream = request.GetRequestStream()) {
							foreach (MARCUpdate update in MainForm.Resources.Values) {
								if (update.UpdateMessage == null) continue;
								stream.Write(Functions.EncodeInt32BigEndian((UInt32)update.UpdateMessage.Length), 0, 4);
								stream.Write(update.UpdateMessage, 0, update.UpdateMessage.Length);
								BytesSent += 4 + update.UpdateMessage.Length; TotalBytesSent += 4 + update.UpdateMessage.Length;
								Exported++; TotalExported++;
							}
						}
					}
					backgroundWorker1.ReportProgress(0, "Exchanging data...");
					using (WebResponse response = request.GetResponse()) {
						backgroundWorker1.ReportProgress(0, "Receiving response...");
						using (Stream stream = response.GetResponseStream()) {
							Byte version = (Byte)stream.ReadByte();
							if (version != 3) throw new Exception("Unsupported server version " + version.ToString());
							BytesReceived++; TotalBytesReceived++;
							for (int numext = stream.ReadByte(); numext > 0; numext--) {
								Byte[] buffer = RealRead(stream, 3, true);
								Byte extid = buffer[0];
								UInt16 extlen = Functions.DecodeInt16BigEndian(buffer, 1);
								RealRead(stream, extlen, false);
								BytesReceived += 3 + extlen; TotalBytesReceived += 3 + extlen;
							}
							while (!backgroundWorker1.CancellationPending) {
								Byte[] buffer = RealRead(stream, 4, false);
								if (buffer == null) break;
								if (backgroundWorker1.CancellationPending) break;
								int len = (int)Functions.DecodeInt32BigEndian(buffer, 0);
								buffer = RealRead(stream, len, true);
								if (backgroundWorker1.CancellationPending) break;
								BytesReceived += len + 4; TotalBytesReceived += len + 4;
								Received++; TotalReceived++;
								try {
									if (MainForm.ImportUpdate(buffer, false)) { Imported++; TotalImported++; }
								} catch (Exception ex) {
									Console.Error.WriteLine(ex.ToString());
								}
								backgroundWorker1.ReportProgress(0, null);
							}
						}
					}
				} catch (Exception ex) {
					Console.Error.WriteLine(ex.ToString());
					backgroundWorker1.ReportProgress(0, ex);
				}
			}
		}

		private void button1_Click(object sender, EventArgs e) {
			if (backgroundWorker1.IsBusy) {
				lblActivity.Text = "Aborting...";
				backgroundWorker1.CancelAsync();
			} else {
				Close();
			}
		}

		private void frmSync_FormClosing(object sender, FormClosingEventArgs e) {
			if (backgroundWorker1.IsBusy) {
				e.Cancel = true;
				lblActivity.Text = "Aborting...";
				backgroundWorker1.CancelAsync();
			}
		}

		private void frmSync_FormClosed(object sender, FormClosedEventArgs e) {
			MainForm.lstResources.Sorted = true;
			MainForm.lstResources.EndUpdate();
		}

		private void backgroundWorker1_ProgressChanged(object sender, ProgressChangedEventArgs e) {
			if (e.ProgressPercentage == 1) {
				lblAddress.Text = (String)e.UserState;
			} else if (e.UserState is Exception) {
				MessageBox.Show(this, "Error: " + ((Exception)e.UserState).Message, "MARC - Synchronization error", MessageBoxButtons.OK, MessageBoxIcon.Error);
			} else if (e.UserState is String) {
				lblActivity.Text = (String)e.UserState;
			} else {
				lblActivity.Text = String.Format("Busy, exported {0}, received {1}, imported {2}, sent {3} kB, received {4} kB...", Exported, Received, Imported, BytesSent / 1000, BytesReceived / 1000);
				lblTotal.Text = String.Format("Total exported {0}, received {1}, imported {2}, sent {3} kB, received {4} kB...", TotalExported, TotalReceived, TotalImported, TotalBytesSent / 1000, TotalBytesReceived / 1000);
			}
		}

		private void backgroundWorker1_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e) {
			lblActivity.Text = "Finished.";
			Close();
		}

		private void frmSync_Load(object sender, EventArgs e) {
			lblActivity.Text = "Starting...";
			MainForm.lstResources.BeginUpdate();
			MainForm.lstResources.Sorted = false;
			backgroundWorker1.RunWorkerAsync();
		}
	}
}
